import csv
import re
from copy import copy
from django.core.management.base import CommandParser
from jutil.command import SafeCommand



def is_int(x) -> bool:
    try:
        int(x)
        return True
    except Exception:
        return False


def dk_iban_load_map(filename: str) -> list:
    """
    Loads Denmark monetary institution codes in CSV format.
    :param filename: CSV file name of the BIC definitions.
    Columns: 4-digit code, bank name, ...
    :return: list of (code, name)
    """
    data_list = []
    name_repl = {
    }
    with open(filename) as fp:
        for row in csv.reader(fp):
            if len(row) >= 2 and is_int(row[0]) and row[1]:
                data_list.append((row[0], row[1]))
    return data_list


class Command(SafeCommand):
    help = 'Generates Python file with Denmark bank info as constants'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--filename', type=str)
        parser.add_argument('--php', action='store_true')

    def do(self, *args, **kw):
        new_bank_list = dk_iban_load_map(kw['filename']) if kw['filename'] else []
        # pprint(bank_list)

        from jutil.bank_const_dk import DK_BANK_CLEARING_MAP
        bank_data = dict(copy(DK_BANK_CLEARING_MAP))
        for code, name in new_bank_list:
            if code not in bank_data:
                bank_data[code] = name

        if kw['php']:
            print('<?php')
            print('')
            print('global $DK_BANK_CLEARING_MAP;')
            print('$DK_BANK_CLEARING_MAP = array(')
            errors = False
            for code, name in bank_data.items():
                print("    '{}' => '{}',".format(code, name))
            print(');')
            if errors:
                print('')
                print('// TODO: fix errors from above marked with "?"')
            print('')
        else:
            print('DK_BANK_CLEARING_MAP = {  # ' + str(len(bank_data)))
            errors = False
            for code, name in bank_data.items():
                print("    '{}': '{}',".format(code, name))
            print('}')
            if errors:
                print('')
                print('# TODO: fix errors from above marked with "?"')
            print('')



import re
from datetime import timedelta
from decimal import Decimal


def format_full_name(first_name: str, last_name: str, max_length: int = 20):
    """
    Limits name length to specified length. Tries to keep name as human-readable an natural as possible.
    :param first_name: First name
    :param last_name: Last name
    :param max_length: Maximum length
    :return: Full name of shortened version depending on length
    """
    # dont allow commas in limited names
    first_name = first_name.replace(',', ' ')
    last_name = last_name.replace(',', ' ')

    # accept short full names as is
    original_full_name = first_name + ' ' + last_name
    if len(original_full_name) <= max_length:
        return original_full_name

    # drop middle names
    first_name = first_name.split(' ')[0]
    full_name = first_name + ' ' + last_name
    if len(full_name) <= max_length:
        return full_name

    # drop latter parts of combined first names
    first_name = re.split(r'[\s\-]', first_name)[0]
    full_name = first_name + ' ' + last_name
    if len(full_name) <= max_length:
        return full_name

    # drop latter parts of multi part last names
    last_name = re.split(r'[\s\-]', last_name)[0]
    full_name = first_name + ' ' + last_name
    if len(full_name) <= max_length:
        return full_name

    # shorten last name to one letter
    last_name = last_name[:1]

    full_name = first_name + ' ' + last_name
    if len(full_name) > max_length:
        raise Exception('Failed to shorten name {}'.format(original_full_name))
    return full_name


def format_timedelta(dt: timedelta) -> str:
    """
    Formats timedelta to readable format, e.g. 1h30min.
    :param dt: timedelta
    :return: str
    """
    seconds = int(dt.total_seconds())
    days, remainder = divmod(seconds, 86400)
    hours, remainder = divmod(remainder, 3600)
    minutes, seconds = divmod(remainder, 60)
    s = ""
    if days > 0:
        s += str(days) + "d"
    if hours > 0:
        s += str(hours) + "h"
    if minutes > 0:
        s += str(minutes) + "min"
    if s == "":
        s = "0min"
    return s


def format_xml(xml_str: str, exceptions: bool = False) -> str:
    """
    Formats XML document as human-readable plain text.
    :param xml_str: XML data as str
    :param exceptions: Raise exceptions on error
    :return: str (Formatted XML str)
    """
    try:
        import xml.dom.minidom
        return xml.dom.minidom.parseString(xml_str).toprettyxml()
    except Exception:
        if exceptions:
            raise
        return xml_str


def format_xml_file(full_path: str, encoding: str = 'UTF-8', exceptions: bool = False, xmllint_path: str = '/usr/bin/xmllint') -> bytes:
    """
    Formats XML file as human-readable plain text and returns result in bytes.
    Tries to format XML file first, if formatting fails the file content is returned as is.
    If the file does not exist empty bytes is returned.
    :param full_path: Full path to XML file
    :param encoding: XML file encoding
    :param exceptions: Raise exceptions on error
    :param xmllint_path: Path to xmllint command line tool (optional). Pass empty string if the tool is not available.
    :return: bytes
    """
    import subprocess
    import xml.dom.minidom
    try:
        if xmllint_path:
            return subprocess.check_output([xmllint_path, '--format', full_path])
        with open(full_path, 'rb') as fp:
            return xml.dom.minidom.parse(fp).toprettyxml(encoding=encoding)
    except Exception as e:
        if exceptions:
            raise
    try:
        with open(full_path, 'rb') as fp:
            return fp.read()
    except Exception as e:
        pass
    return b''


def dec1(a) -> Decimal:
    """
    Converts number to Decimal with 1 decimal digits.
    :param a: Number
    :return: Decimal with 1 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.0'))


def dec2(a) -> Decimal:
    """
    Converts number to Decimal with 2 decimal digits.
    :param a: Number
    :return: Decimal with 2 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.00'))


def dec3(a) -> Decimal:
    """
    Converts number to Decimal with 3 decimal digits.
    :param a: Number
    :return: Decimal with 3 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.000'))


def dec4(a) -> Decimal:
    """
    Converts number to Decimal with 4 decimal digits.
    :param a: Number
    :return: Decimal with 4 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.0000'))


def dec5(a) -> Decimal:
    """
    Converts number to Decimal with 5 decimal digits.
    :param a: Number
    :return: Decimal with 4 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.00000'))


def dec6(a) -> Decimal:
    """
    Converts number to Decimal with 6 decimal digits.
    :param a: Number
    :return: Decimal with 4 decimal digits
    """
    return Decimal(a).quantize(Decimal('1.000000'))



from django.core.management.base import BaseCommand, CommandParser
from jutil.format import format_xml_file


class Command(BaseCommand):
    help = 'Formats XML file'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('filename', type=str)
        parser.add_argument('--encoding', type=str, default='UTF-8')
        parser.add_argument('--xmllint-path', type=str, default='/usr/bin/xmllint')

    def handle(self, **options):
        filename = options['filename']
        encoding = options['encoding']
        xmllint_path = options['xmllint_path']
        print(format_xml_file(filename, encoding=encoding, xmllint_path=xmllint_path).decode())



import re
import unicodedata
from datetime import date
from decimal import Decimal
from random import randint

from django.core.exceptions import ValidationError
from django.utils.timezone import now
from django.utils.translation import ugettext as _
from jutil.bank_const_iban import IBAN_LENGTH_BY_COUNTRY


EMAIL_VALIDATOR = re.compile(r'[a-zA-Z0-9\._-]+@[a-zA-Z0-9\._-]+\.[a-zA-Z]+')
PHONE_FILTER = re.compile(r'[^+0-9]')
PHONE_VALIDATOR = re.compile(r'\+?\d{6,}')
PASSPORT_FILTER = re.compile(r'[^-A-Z0-9]')
STRIP_NON_NUMBERS = re.compile(r'[^0-9]')
STRIP_NON_ALPHANUMERIC = re.compile(r'[^0-9A-Za-z]')
STRIP_WHITESPACE = re.compile(r'\s+')
IBAN_FILTER = re.compile(r'[^A-Z0-9]')
DIGIT_FILTER = re.compile(r'[^0-9]')


def phone_filter(v: str) -> str:
    return PHONE_FILTER.sub('', str(v)) if v else ''


def email_filter(v: str) -> str:
    return str(v).lower().strip() if v else ''


def email_validator(v: str) -> str:
    v = email_filter(v)
    if not v or not EMAIL_VALIDATOR.fullmatch(v):
        v_str = _('Missing value') if not v else str(v)
        raise ValidationError(_('Invalid email') + ': {}'.format(v_str), code='invalid_email')


def phone_validator(v: str):
    v = phone_filter(v)
    if not v or not PHONE_VALIDATOR.fullmatch(v):
        v_str = _('Missing value') if v is None else str(v)
        raise ValidationError(_('Invalid phone number') + ': {}'.format(v_str), code='invalid_phone')


def passport_filter(v: str) -> str:
    return PASSPORT_FILTER.sub('', str(v).upper()) if v else ''


def passport_validator(v: str):
    v = passport_filter(v)
    if not v or len(v) < 5:
        v_str = _('Missing value') if v is None else str(v)
        raise ValidationError(_('Invalid passport number') + ': {}'.format(v_str), code='invalid_passport')


def ascii_filter(v: str) -> str:
    """
    Replaces Unicode accent characters with plain ASCII.
    For example remove_accents('HELÉN') == 'HELEN'.
    :param v: str
    :return: str
    """
    return unicodedata.normalize('NFKD', v).encode('ASCII', 'ignore').decode()


def digit_filter(v: str) -> str:
    return DIGIT_FILTER.sub('', str(v)) if v else ''


def iban_filter(v: str) -> str:
    return IBAN_FILTER.sub('', str(v).upper()) if v else ''


def iban_filter_readable(acct) -> str:
    acct = iban_filter(acct)
    if acct:
        i = 0
        j = 4
        out = ''
        nlen = len(acct)
        while i < nlen:
            if out:
                out += ' '
            out += acct[i:j]
            i = j
            j += 4
        return out
    return acct


def iban_validator(v: str):
    # validate IBAN numeric part
    v = iban_filter(v)
    if not v:
        raise ValidationError(_('Invalid IBAN account number') + ': {}'.format(_('Missing value')), code='invalid_iban')
    digits = '0123456789'
    num = ''
    for ch in v[4:] + v[0:4]:
        if ch not in digits:
            ch = str(ord(ch) - ord('A') + 10)
        num += ch
    x = Decimal(num) % Decimal(97)
    if x != Decimal(1):
        raise ValidationError(_('Invalid IBAN account number') + ': {}'.format(v), code='invalid_iban')

    # validate prefix and length
    country = v[:2]
    iban_len = IBAN_LENGTH_BY_COUNTRY.get(country, 0)
    if iban_len != len(v):
        raise ValidationError(_('Invalid IBAN account number') + ' ({}.1): {}'.format(country, v), code='invalid_iban')


def validate_country_iban(v: str, country: str, length: int):
    v = iban_filter(v)
    if v[0:2] != country:
        raise ValidationError(_('Invalid IBAN account number') + ' ({}.2): {}'.format(country, v), code='invalid_iban')
    iban_validator(v)


def iban_bank_info(v: str) -> (str, str):
    """
    Returns BIC code and bank name from IBAN number.
    :param v: IBAN account number
    :return: (BIC code, bank name) or ('', '') if not found / unsupported country
    """
    v = iban_filter(v)
    prefix = v[:2]
    func_name = prefix.lower() + '_iban_bank_info'  # e.g. fi_iban_bank_info, be_iban_bank_info
    func = globals().get(func_name)
    if func is not None:
        return func(v)
    else:
        return '', ''


def iban_bic(v: str) -> str:
    """
    Returns BIC code from IBAN number.
    :param v: IBAN account number
    :return: BIC code or '' if not found
    """
    info = iban_bank_info(v)
    return info[0] if info else ''


def calculate_age(born: date, today: date or None=None) -> int:
    if not today:
        today = now().date()
    return today.year - born.year - ((today.month, today.day) < (born.month, born.day))


def filter_country_company_reg_id(country_code: str, v: str):
    if country_code == 'FI':
        return fi_company_reg_id_filter(v)
    return PASSPORT_FILTER.sub('', v)


def validate_country_company_reg_id(country_code: str, v: str):
    if country_code == 'FI':
        fi_company_reg_id_validator(v)


# ============================================================================
# Country specific functions (countries in alphabetical order)
# ============================================================================


# ----------------------------------------------------------------------------
# Belgium
# ----------------------------------------------------------------------------

def be_iban_validator(v: str):
    validate_country_iban(v, 'BE', IBAN_LENGTH_BY_COUNTRY.get('BE'))


def be_iban_bank_info(v: str) -> (str, str):
    """
    Returns BIC code and bank name from BE IBAN number.
    :param v: IBAN account number
    :return: (BIC code, bank name) or ('', '') if not found
    """
    from jutil.bank_const_be import BE_BIC_BY_ACCOUNT_NUMBER, BE_BANK_NAME_BY_BIC
    v = iban_filter(v)
    bic = BE_BIC_BY_ACCOUNT_NUMBER.get(v[4:7], None)
    return (bic, BE_BANK_NAME_BY_BIC[bic]) if bic is not None else ('', '')


# ----------------------------------------------------------------------------
# Denmark
# ----------------------------------------------------------------------------

def dk_iban_validator(v: str):
    validate_country_iban(v, 'DK', IBAN_LENGTH_BY_COUNTRY.get('DK'))


def dk_clearing_code_bank_name(v: str) -> str:
    from jutil.bank_const_dk import DK_BANK_CLEARING_MAP
    v = iban_filter(v)
    if v.startswith('DK'):
        v = v[4:]
    return DK_BANK_CLEARING_MAP.get(v[:4], '')


def dk_iban_bank_info(v: str) -> (str, str):
    """
    Returns empty string (BIC not available) and bank name from DK IBAN number.
    DK5000400440116243
    :param v: IBAN account number
    :return: ('', bank name) or ('', '') if not found
    """
    return '', dk_clearing_code_bank_name(v)


# ----------------------------------------------------------------------------
# Estonia
# ----------------------------------------------------------------------------

def ee_iban_validator(v: str):
    validate_country_iban(v, 'EE', IBAN_LENGTH_BY_COUNTRY.get('EE'))


# ----------------------------------------------------------------------------
# Finland
# ----------------------------------------------------------------------------

FI_SSN_FILTER = re.compile(r'[^-A-Z0-9]')
FI_SSN_VALIDATOR = re.compile(r'^\d{6}[+-A]\d{3}[\d\w]$')
FI_COMPANY_REG_ID_FILTER = re.compile(r'[^0-9]')


def fi_payment_reference_number(num: str):
    """
    Appends Finland reference number checksum to existing number.
    :param num: At least 3 digits
    :return: Number plus checksum
    """
    assert isinstance(num, str)
    num = STRIP_WHITESPACE.sub('', num)
    num = re.sub(r'^0+', '', num)
    assert len(num) >= 3
    weights = [7, 3, 1]
    weighed_sum = 0
    numlen = len(num)
    for j in range(numlen):
        weighed_sum += int(num[numlen - 1 - j]) * weights[j % 3]
    return num + str((10 - (weighed_sum % 10)) % 10)


def fi_payment_reference_validator(v: str):
    v = STRIP_WHITESPACE.sub('', v)
    if fi_payment_reference_number(v[:-1]) != v:
        raise ValidationError(_('Invalid payment reference: {}').format(v))


def iso_payment_reference_validator(v: str):
    """
    Validates ISO reference number checksum.
    :param v: Reference number
    """
    num = ''
    v = STRIP_WHITESPACE.sub('', v)
    for ch in v[4:] + v[0:4]:
        x = ord(ch)
        if ord('0') <= x <= ord('9'):
            num += ch
        else:
            x -= 55
            if x < 10 or x > 35:
                raise ValidationError(_('Invalid payment reference: {}').format(v))
            num += str(x)
    res = Decimal(num) % Decimal('97')
    if res != Decimal('1'):
        raise ValidationError(_('Invalid payment reference: {}').format(v))


def fi_iban_validator(v: str):
    validate_country_iban(v, 'FI', IBAN_LENGTH_BY_COUNTRY.get('FI'))


def fi_iban_bank_info(v: str) -> (str, str):
    """
    Returns BIC code and bank name from FI IBAN number.
    :param v: IBAN account number
    :return: (BIC code, bank name) or ('', '') if not found
    """
    from jutil.bank_const_fi import FI_BIC_BY_ACCOUNT_NUMBER, FI_BANK_NAME_BY_BIC
    v = iban_filter(v)
    bic = FI_BIC_BY_ACCOUNT_NUMBER.get(v[4:7], None)
    return (bic, FI_BANK_NAME_BY_BIC[bic]) if bic is not None else ('', '')


def fi_ssn_filter(v: str) -> str:
    return FI_SSN_FILTER.sub('', v.upper())


def fi_company_reg_id_filter(v: str) -> str:
    v = FI_COMPANY_REG_ID_FILTER.sub('', v)
    return v[:-1] + '-' + v[-1:] if len(v) >= 2 else ''


def fi_company_reg_id_validator(v0: str) -> str:
    v = fi_company_reg_id_filter(v0)
    prefix = v[:2]
    if v[-2:-1] != '-' and prefix != 'FI':
        raise ValidationError(_('Invalid company registration ID')+' (FI.1): {}'.format(v0), code='invalid_company_reg_id')
    v = v.replace('-', '', 1)
    if len(v) != 8:
        raise ValidationError(_('Invalid company registration ID')+' (FI.2): {}'.format(v0), code='invalid_company_reg_id')
    multipliers = (7, 9, 10, 5, 8, 4, 2)
    x = 0
    for i, m in enumerate(multipliers):
        x += int(v[i]) * m
    quotient, remainder = divmod(x, 11)
    if remainder == 1:
        raise ValidationError(_('Invalid company registration ID')+' (FI.3): {}'.format(v0), code='invalid_company_reg_id')
    if remainder >= 2:
        check_digit = str(11 - remainder)
        if check_digit != v[-1:]:
            raise ValidationError(_('Invalid company registration ID')+' (FI.4): {}'.format(v0), code='invalid_company_reg_id')


def fi_ssn_validator(v: str):
    v = fi_ssn_filter(v)
    if not FI_SSN_VALIDATOR.fullmatch(v):
        raise ValidationError(_('Invalid personal identification number')+' (FI.1): {}'.format(v), code='invalid_ssn')
    d = int(Decimal(v[0:6] + v[7:10]) % Decimal(31))
    digits = {
        10: 'A', 	11: 'B', 	12: 'C', 	13: 'D', 	14: 'E', 	15: 'F', 	16: 'H',
        17: 'J', 	18: 'K', 	19: 'L', 	20: 'M', 	21: 'N', 	22: 'P', 	23: 'R',
        24: 'S', 	25: 'T', 	26: 'U', 	27: 'V', 	28: 'W', 	29: 'X', 	30: 'Y',
    }
    ch = str(d)
    if d in digits:
        ch = digits[d]
    if ch != v[-1:]:
        raise ValidationError(_('Invalid personal identification number')+' (FI.2): {}'.format(v), code='invalid_ssn')


def fi_ssn_generator():
    day = randint(1, 28)
    month = randint(1, 12)
    year = randint(1920, 1999)
    suffix = randint(100, 999)
    v = '{:02}{:02}{:02}-{}'.format(day, month, year-1900, suffix)
    d = int(Decimal(v[0:6] + v[7:10]) % Decimal(31))
    digits = {
        10: 'A', 	11: 'B', 	12: 'C', 	13: 'D', 	14: 'E', 	15: 'F', 	16: 'H',
        17: 'J', 	18: 'K', 	19: 'L', 	20: 'M', 	21: 'N', 	22: 'P', 	23: 'R',
        24: 'S', 	25: 'T', 	26: 'U', 	27: 'V', 	28: 'W', 	29: 'X', 	30: 'Y',
    }
    ch = str(d)
    if d in digits:
        ch = digits[d]
    return v + ch


def fi_ssn_birthday(v: str) -> date:
    v = fi_ssn_filter(v)
    fi_ssn_validator(v)
    sep = v[6]  # 231298-965X
    year = int(v[4:6])
    month = int(v[2:4])
    day = int(v[0:2])
    if sep == '+':  # 1800
        year += 1800
    elif sep == '-':
        year += 1900
    elif sep == 'A':
        year += 2000
    return date(year, month, day)


def fi_ssn_age(ssn: str, today: date or None=None) -> int:
    return calculate_age(fi_ssn_birthday(ssn), today)


# ----------------------------------------------------------------------------
# Sweden
# ----------------------------------------------------------------------------

SE_SSN_FILTER = re.compile(r'[^-0-9]')
SE_SSN_VALIDATOR = re.compile(r'^\d{6}[-]\d{3}[\d]$')


def se_iban_validator(v: str):
    validate_country_iban(v, 'SE', IBAN_LENGTH_BY_COUNTRY.get('SE'))


def se_ssn_filter(v: str) -> str:
    return SE_SSN_FILTER.sub('', v.upper())


def se_ssn_validator(v: str):
    v = se_ssn_filter(v)
    if not SE_SSN_VALIDATOR.fullmatch(v):
        raise ValidationError(_('Invalid personal identification number')+' (SE.1): {}'.format(v), code='invalid_ssn')
    v = STRIP_NON_NUMBERS.sub('', v)
    dsum = 0
    for i in range(9):
        x = int(v[i])
        if i & 1 == 0:
            x += x
        # print('summing', v[i], 'as', x)
        xsum = x % 10 + int(x/10) % 10
        # print(v[i], 'xsum', xsum)
        dsum += xsum
    # print('sum', dsum)
    rem = dsum % 10
    # print('rem', rem)
    checksum = 10 - rem
    if checksum == 10:
        checksum = 0
    # print('checksum', checksum)
    if int(v[-1:]) != checksum:
        raise ValidationError(_('Invalid personal identification number')+' (SE.2): {}'.format(v), code='invalid_ssn')


def se_clearing_code_bank_info(account_number: str) -> (str, int):
    """
    Returns Sweden bank info by clearning code.
    :param account_number: Swedish account number with clearing code as prefix
    :return: (Bank name, account digit count) or ('', None) if not found
    """
    from jutil.bank_const_se import SE_BANK_CLEARING_LIST
    v = digit_filter(account_number)
    clearing = v[:4]
    for name, begin, end, acc_digits in SE_BANK_CLEARING_LIST:
        if begin <= clearing <= end:
            return name, acc_digits
    return '', None



from datetime import datetime
from django.utils.translation import ugettext as _
from rest_framework.exceptions import ValidationError
import pytz
from dateutil.parser import parse as dateutil_parse


TRUE_VALUES = (
    'true',
    '1',
    'yes',
)

FALSE_VALUES = (
    'none',
    'null',
    'false',
    '0',
    'no',
)


def parse_bool(v, default = None, exceptions: bool = True) -> bool:
    """
    Parses boolean value
    :param v: Input string
    :param default: Default value if exceptions=False
    :param exceptions: Raise exception on error or not
    :return: bool
    """
    if isinstance(v, bool):
        return v
    s = str(v).lower()
    if s in TRUE_VALUES:
        return True
    elif s in FALSE_VALUES:
        return False
    else:
        if exceptions:
            msg = _("%(value)s is not one of the available choices") % {'value': v}
            raise ValidationError(msg)
        return default


def parse_datetime(v: str, default = None, tz = None, exceptions: bool = True) -> datetime:
    """
    Parses str to timezone-aware datetime.
    :param v: Input string
    :param default: Default value if exceptions=False
    :param tz: Default pytz timezone or None if use utc as default
    :param exceptions: Raise exception on error or not
    :return: datetime
    """
    try:
        t = dateutil_parse(v, default=datetime(2000, 1, 1))
        if tz is None:
            tz = pytz.utc
        return t if t.tzinfo else tz.localize(t)
    except Exception:
        if exceptions:
            msg = _("'%(value)s' value has an invalid format. It must be in YYYY-MM-DD HH:MM[:ss[.uuuuuu]][TZ] format.") % {'value': v}
            raise ValidationError(msg)
        return default



import logging


logger = logging.getLogger(__name__)


def send_email(recipients: list, subject: str, text: str, html: str='', sender: str='', files: list=[], exceptions: bool=False):
    """
    :param recipients: List of recipients; or single email (str); or comma-separated email list (str); or list of name-email pairs (e.g. settings.ADMINS)
    :param subject: Subject of the email
    :param text: Body (text)
    :param html: Body (html)
    :param sender: Sender email, or settings.DEFAULT_FROM_EMAIL if missing
    :param files: Paths to files to attach
    :param exceptions: Raise exception if email sending fails
    :return: Status code 202 if all emails were sent successfully, error status code otherwise
    """
    import sendgrid
    from sendgrid.helpers.mail import Content, Mail, Attachment
    from django.conf import settings
    from base64 import b64encode
    from os.path import basename
    from django.utils.timezone import now
    from jutil.logs import log_event

    try:
        # default sender to settings.DEFAULT_FROM_EMAIL
        if not sender:
            sender = settings.DEFAULT_FROM_EMAIL

        # support multiple recipient list styles
        if isinstance(recipients, str):  # allow single email and comma-separated list as input
            recipients = [str(r).strip() for r in recipients.split(',')]

        sg = sendgrid.SendGridAPIClient(apikey=settings.EMAIL_SENDGRID_API_KEY)
        from_email = sendgrid.Email(sender or settings.DEFAULT_FROM_EMAIL)
        content = Content('text/plain', text) if not html else Content('text/html', html)

        attachments = []
        for filename in files:
            with open(filename, 'rb') as fp:
                attachment = Attachment()
                attachment.content = b64encode(fp.read()).decode()
                attachment.type = "application/octet-stream"
                attachment.filename = basename(filename)
                attachment.content_id = basename(filename)
                attachment.disposition = "attachment"
                attachments.append(attachment)
    except Exception as e:
        logger.error(e)
        if exceptions:
            raise
        return -1

    status_codes = []
    for recipient in recipients:
        try:
            t = now()

            to_email = sendgrid.Email()
            if isinstance(recipient, str):
                to_email.email = recipient
            elif (isinstance(recipient, list) or isinstance(recipient, tuple)) and len(recipient) == 2:
                to_email.name = recipient[0]
                to_email.email = recipient[1]
            else:
                raise Exception('Invalid recipient format: {}'.format(recipient))

            mail = Mail(from_email=from_email, subject=subject, to_email=to_email, content=content)
            for attachment in attachments:
                mail.add_attachment(attachment)
            res = sg.client.mail.send.post(request_body=mail.get())

            send_dt = (now()-t).total_seconds()
            if res.status_code == 202:
                log_event('EMAIL_SENT', data={'time': send_dt, 'to': recipient, 'subject': subject, 'status': res.status_code})
            else:
                log_event('EMAIL_ERROR', data={'time': send_dt, 'to': recipient, 'subject': subject, 'status': res.status_code, 'body': res.body})

            status_codes.append(res.status_code)
        except Exception as e:
            logger.error(e)
            if exceptions:
                raise
            status_codes.append(-1)

    for status in status_codes:
        if status != 202:
            return status
    return 202



def _split_obj_attr_path(obj, key: str, exceptions: bool = True) -> (str, str):
    obj0 = obj
    key0 = key
    key_parts = key.split('.')
    key_path = key_parts[:-1]
    key_name = key_parts[-1]
    for next_obj in key_path:
        if not hasattr(obj, next_obj):
            if exceptions:
                raise AttributeError('{} not in {}'.format(key0, obj0))
            return None, ''
        obj = getattr(obj, next_obj)
    return obj, key_name


def set_obj_attr(obj, key: str, val):
    """
    Set object property. Support '.' separate path to sub-objects, for example
    set_key_value(user, 'profile.address', 'Lapinrinne 1') sets user.profile.address as 'Lapinrinne 1'.
    :param obj: Object
    :param key: Attribute name
    :param val: New attribute value
    :return: None
    """
    obj, key_name = _split_obj_attr_path(obj, key)
    setattr(obj, key_name, val)


def get_obj_attr(obj, key: str, default = None, exceptions: bool = True):
    """
    Get object property. Support '.' separate path to sub-objects, for example
    get_key_value(user, 'profile.address') gets user.profile.address field value.
    :param obj: Object
    :param key: Attribute name
    :param default: Default return value if exceptions=False
    :param exceptions: Raise AttributeError or not. Default is True.
    :return: Attribute value
    """
    obj, key_name = _split_obj_attr_path(obj, key, exceptions=exceptions)
    return getattr(obj, key_name) if obj is not None else default



from pprint import pprint
from django.core.exceptions import ValidationError
from django.core.management.base import CommandParser
from jutil.command import SafeCommand
from jutil.dict import sorted_dict
from jutil.bank_const_fi import FI_BIC_BY_ACCOUNT_NUMBER, FI_BANK_NAME_BY_BIC


def fi_iban_load_map(filename: str) -> dict:
    """
    Loads Finnish monetary institution codes and BICs in CSV format.
    Map which is based on 3 digits as in FIXX<3 digits>.
    Can be used to map Finnish IBAN number to bank information.
    Format: dict('<3 digits': (BIC, name), ...)
    :param filename: CSV file name of the BIC definitions. Columns: National ID, BIC Code, Institution Name
    """
    out = {}
    with open(filename, 'rt') as fp:
        lines = [line.strip().split(',') for line in fp.readlines()]
        ver = lines.pop(0)
        head = lines.pop(0)
        if head != ['National ID', 'BIC Code', 'Financial Institution Name']:
            raise ValidationError('Incompatible file content in {}'.format(filename))
        for line in lines:
            if len(line) == 3 and line[0]:
                nat_id = str(line[0]).strip()
                bic_code = line[1].strip()
                name = line[2].strip()
                out[nat_id] = (bic_code, name)
    return out


class Command(SafeCommand):
    help = 'Generates Python file with Finnish bank info as constants'

    def add_arguments(self, parser: CommandParser):
        parser.add_argument('--filename', type=str)
        parser.add_argument('--php', action='store_true')

    def do(self, *args, **kw):
        iban_map = fi_iban_load_map(kw['filename']) if kw['filename'] else {}
        bic_by_acc = FI_BIC_BY_ACCOUNT_NUMBER
        for acc, bank in iban_map.items():
            bic_by_acc[acc] = bank[0]
        bic_map = FI_BANK_NAME_BY_BIC
        for acc, bank in iban_map.items():
            if bank[0] not in bic_map:
                bic_map[bank[0]] = bank[1]
        bic_by_acc = sorted_dict(bic_by_acc)
        bic_map = sorted_dict(bic_map)

        if kw['php']:
            print('<?php')
            print('')
            print('global $FI_BIC_BY_ACCOUNT_NUMBER;')
            print('$FI_BIC_BY_ACCOUNT_NUMBER = array(')
            for acc, bank in iban_map.items():
                print("    '{}' => '{}',".format(acc, bank[0]))
            print(');')
            print('')
            print('global $FI_BANK_NAME_BY_BIC;')
            print('$FI_BANK_NAME_BY_BIC = array(')
            for bic, bank in bic_map.items():
                print("    '{}' => '{}',".format(bic, bank))
            print(');')
            print('')
        else:
            print('FI_BIC_BY_ACCOUNT_NUMBER = {  # ' + str(len(bic_by_acc.items())))
            for acc, bic in bic_by_acc.items():
                print("    '{}': '{}',".format(acc, bic))
            print('}\n')
            print('FI_BANK_NAME_BY_BIC = {  # ' + str(len(bic_map.items())))
            for bic, name in bic_map.items():
                print("    '{}': '{}',".format(bic, name))
            print('}')
            print('')
